﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;
using Microsoft.Research.Kinect.Audio;
using Microsoft.Speech.AudioFormat;
using Microsoft.Speech.Recognition;
using Microsoft.Speech;
using Actions;

namespace LilyNUI
{
    public class SpeechTracker
    {
        #region Events

        public class IntentionArgs : EventArgs
        {
            public Context context { get; set; }
            public General general { get; set; }
            public Research research { get; set; }
            public Office office { get; set; }
            public Operations operations { get; set; }
            public Meeting meeting { get; set; }
            public Translate translate { get; set; }
            public Entertain entertain { get; set; }
            public string phrase { get; set; }
            public bool contextOnly { get; set; }
        }

        public class SpeechContextChangedArgs : EventArgs
        {
            public Context context { get; set; }
        }

        public event EventHandler<IntentionArgs> SpeechDetected;
        public event EventHandler<SpeechContextChangedArgs> SpeechContextChanged;

        #endregion

        #region Enumerations

        public enum Context
        {
            Context = 0,
            General,
            Research,
            Office,
            Operations,
            Meeting,
            Translate,
            Entertain
        }

        public enum General
        {
            None = 0,
            Time,
            Log,
            Schedule,
            Pulse,
            Shutdown,
            ColorViewer,
            DepthViewer,
            SkeletonViewer,
            Buttons,
            Affirmation,
            Thanks,
            Negative,
            TrackOn,
            TrackOff,
            TranslationView,
            NewsView
        }

        public enum Research
        {
            None=0,
            Lookup,
            Spelling,
            News,
        }

        public enum Office
        {
            None = 0,
            Word,
            OpenADocument,
            Outlook,
            Excel,
            PowerPoint,
            OpenAPresentation,
            PPTNextSlide,
            PPTPreviousSlide,
            StartPresentation,
            OpenAProjectPlan
        }

        public enum Operations
        {
            None = 0,
            SkeletonTracker,
            CameraUp,
            CameraDown,
            CameraCenter,
            CourseControl,
            FineControl
        }

        public enum Meeting
        {
            None = 0,
            Snapshot,
            CaptureVideo,
            CaptureAudio,
            Stop
        }

        public enum Translate
        {
            None = 0,
            TurkishEnglish,
            EnglishTurkish,
            ArabicEnglish,
            EnglishArabic,
            SpanishEnglish,
            EnglishSpanish
        }

        public enum Entertain
        {
            None = 0,
            Music,
            Video
        }

        #endregion

        #region Grammars

        struct Intention
        {
            public Context context;
            public General general;
            public Research research;
            public Office office;
            public Operations kinnyops;
            public Meeting meeting;
            public Translate translate;
            public Entertain entertain;
            public bool contextOnly;
        }

        Dictionary<string, Intention> ContextPhrases = new Dictionary<string, Intention>()
        {
            {"General", new Intention()                                  {context= Context.General, contextOnly= true}},
            {"Research", new Intention()                                 {context= Context.Research, contextOnly= true}},
            {"I need some information", new Intention()                  {context= Context.Research, contextOnly= true}},
            {"Lets Research", new Intention()                            {context= Context.Research, contextOnly= true}},
            {"I need to do some Research", new Intention()               {context= Context.Research, contextOnly= true}},
            {"Lily, I need you to do some Research", new Intention()     {context= Context.Research, contextOnly= true}},
            {"Office", new Intention()                                   {context= Context.Office, contextOnly= true}},
            {"Kinect", new Intention()                                   {context= Context.Operations, contextOnly= true}},
            {"Meeting", new Intention()                                  {context= Context.Meeting, contextOnly= true}},
            {"Translate", new Intention()                                {context= Context.Translate, contextOnly= true}},
            {"Lily can you translate for me", new Intention()            {context= Context.Translate, contextOnly= true}},
            {"Lily please translate for me", new Intention()             {context= Context.Translate, contextOnly= true}},
            {"Translation", new Intention()                              {context= Context.Translate, contextOnly= true}},
            {"Entertain", new Intention()                                {context= Context.Entertain, contextOnly= true}},
        };

        Dictionary<string, Intention> GeneralPhrases = new Dictionary<string, Intention>()
        {
            {"Time", new Intention()                                     {context= Context.General, contextOnly= false, general= General.Time}},
            {"What is the time", new Intention()                         {context= Context.General, contextOnly= false, general= General.Time}},
            {"What time is it", new Intention()                          {context= Context.General, contextOnly= false, general= General.Time}},
            {"Lily what time is it", new Intention()                     {context= Context.General, contextOnly= false, general= General.Time}},
            {"Error log", new Intention()                                {context= Context.General, contextOnly= false, general= General.Log}},
            {"Schedule", new Intention()                                 {context= Context.General, contextOnly= false, general= General.Schedule}},
            {"Lily", new Intention()                                     {context= Context.General, contextOnly= false, general= General.Pulse}},
            {"Lily are you there", new Intention()                       {context= Context.General, contextOnly= false, general= General.Pulse}},
            {"Lily are you listening", new Intention()                   {context= Context.General, contextOnly= false, general= General.Pulse}},
            {"Lily please shut down", new Intention()                    {context= Context.General, contextOnly= false, general= General.Shutdown}},
            {"Color viewer", new Intention()                             {context= Context.General, contextOnly= false, general= General.ColorViewer}},
            {"Depth viewer", new Intention()                             {context= Context.General, contextOnly= false, general= General.DepthViewer}},
            {"Skeleton viewer", new Intention()                          {context= Context.General, contextOnly= false, general= General.SkeletonViewer}},
            {"Lily, show me the color viewer", new Intention()           {context= Context.General, contextOnly= false, general= General.ColorViewer}},
            {"Lily, show me the depth viewer", new Intention()           {context= Context.General, contextOnly= false, general= General.DepthViewer}},
            {"Lily, show me the skeleton viewer", new Intention()        {context= Context.General, contextOnly= false, general= General.SkeletonViewer}},
            {"Translation viewer", new Intention()                       {context= Context.General, contextOnly= false, general= General.TranslationView}},
            {"News viewer", new Intention()                              {context= Context.General, contextOnly= false, general= General.NewsView}},
            {"Show me the buttons", new Intention()                      {context= Context.General, contextOnly= false, general= General.Buttons}},
            {"Lily, buttons please", new Intention()                     {context= Context.General, contextOnly= false, general= General.Buttons}},
            {"Buttons please", new Intention()                           {context= Context.General, contextOnly= false, general= General.Buttons}},
            {"Lily, show me the buttons", new Intention()                {context= Context.General, contextOnly= false, general= General.Buttons}},
            {"Lily, can you show me the buttons", new Intention()        {context= Context.General, contextOnly= false, general= General.Buttons}},
            {"Lily, good job", new Intention()                           {context= Context.General, contextOnly= false, general= General.Affirmation}},
            {"good job Lily", new Intention()                            {context= Context.General, contextOnly= false, general= General.Affirmation}},
            {"awesome Lily", new Intention()                             {context= Context.General, contextOnly= false, general= General.Affirmation}},
            {"awesome job Lily", new Intention()                         {context= Context.General, contextOnly= false, general= General.Affirmation}},
            {"Thank you Lily", new Intention()                           {context= Context.General, contextOnly= false, general= General.Thanks}},
            {"Thanks Lily", new Intention()                              {context= Context.General, contextOnly= false, general= General.Thanks}},
            {"No Lily", new Intention()                                  {context= Context.General, contextOnly= false, general= General.Negative}},
            {"Thats not right Lily", new Intention()                     {context= Context.General, contextOnly= false, general= General.Negative}},
            {"Lily thats not right", new Intention()                     {context= Context.General, contextOnly= false, general= General.Negative}},
            {"Thats not what I meant Lily", new Intention()              {context= Context.General, contextOnly= false, general= General.Negative}},
            {"That sucks Lily", new Intention()                          {context= Context.General, contextOnly= false, general= General.Negative}},
            {"Lily track off", new Intention()                           {context= Context.General, contextOnly= false, general= General.TrackOff}},
            {"Lily please stop tracking", new Intention()                {context= Context.General, contextOnly= false, general= General.TrackOff}},
            {"Lily turn tracking off", new Intention()                   {context= Context.General, contextOnly= false, general= General.TrackOff}},
            {"Lily please turn tracking off", new Intention()            {context= Context.General, contextOnly= false, general= General.TrackOff}},
            {"Lily track on", new Intention()                            {context= Context.General, contextOnly= false, general= General.TrackOn}},
            {"Lily please start tracking", new Intention()               {context= Context.General, contextOnly= false, general= General.TrackOn}},
            {"Lily turn tracking on", new Intention()                    {context= Context.General, contextOnly= false, general= General.TrackOn}},
            {"Lily please turn tracking on", new Intention()             {context= Context.General, contextOnly= false, general= General.TrackOn}},
        };

        Dictionary<string, Intention> ResearchPhrases = new Dictionary<string, Intention>()
        {
            {"Look up", new Intention()                        {context= Context.Research, contextOnly= false, research = Research.Lookup}},
            {"Bing", new Intention()                           {context= Context.Research, contextOnly= false, research = Research.Lookup}},
            {"Google", new Intention()                         {context= Context.Research, contextOnly= false, research = Research.Lookup}},
            {"Search for", new Intention()                     {context= Context.Research, contextOnly= false, research = Research.Lookup}},
            {"Spell", new Intention()                          {context= Context.Research, contextOnly= false, research = Research.Spelling}},
            {"Checking the spelling of", new Intention()       {context= Context.Research, contextOnly= false, research = Research.Spelling}},
            {"How do you spell", new Intention()               {context= Context.Research, contextOnly= false, research = Research.Spelling}},
            {"News", new Intention()                           {context= Context.Research, contextOnly= false, research = Research.News}},
            {"What's in the headlines", new Intention()        {context= Context.Research, contextOnly= false, research = Research.News}},
            {"What are the headlines", new Intention()         {context= Context.Research, contextOnly= false, research = Research.News}},
            {"What's in the news", new Intention()             {context= Context.Research, contextOnly= false, research = Research.News}},
            {"Lily what's in the news", new Intention()        {context= Context.Research, contextOnly= false, research = Research.News}},
            {"What's making the news", new Intention()         {context= Context.Research, contextOnly= false, research = Research.News}},
            {"Can you show me the news", new Intention()       {context= Context.Research, contextOnly= false, research = Research.News}},
            {"Lily please show me the news", new Intention()   {context= Context.Research, contextOnly= false, research = Research.News}},
        };

        Dictionary<string, Intention> OfficePhrases = new Dictionary<string, Intention>()
        {
            {"Excel", new Intention()                          {context= Context.Office, contextOnly= false, office = Office.Excel}},
            {"Lets crunch some numbers", new Intention()       {context= Context.Office, contextOnly= false, office = Office.Excel}},
            {"Start a spreadsheet", new Intention()            {context= Context.Office, contextOnly= false, office = Office.Excel}},
            {"Create a spreadsheet", new Intention()           {context= Context.Office, contextOnly= false, office = Office.Excel}},
            {"Open Excel", new Intention()                     {context= Context.Office, contextOnly= false, office = Office.Excel}},
            {"Outlook", new Intention()                        {context= Context.Office, contextOnly= false, office = Office.Outlook}},
            {"Open Outlook", new Intention()                   {context= Context.Office, contextOnly= false, office = Office.Outlook}},
            {"Check email", new Intention()                    {context= Context.Office, contextOnly= false, office = Office.Outlook}},
            {"Lets look at email", new Intention()             {context= Context.Office, contextOnly= false, office = Office.Outlook}},
            {"Show me my email", new Intention()               {context= Context.Office, contextOnly= false, office = Office.Outlook}},
            {"Show me Outlook", new Intention()                {context= Context.Office, contextOnly= false, office = Office.Outlook}},
            {"Power point", new Intention()                    {context= Context.Office, contextOnly= false, office = Office.PowerPoint}},
            {"Open Power point", new Intention()               {context= Context.Office, contextOnly= false, office = Office.PowerPoint}},
            {"Open the presentation", new Intention()          {context= Context.Office, contextOnly= false, office = Office.OpenAPresentation}},
            {"Create a Power point", new Intention()           {context= Context.Office, contextOnly= false, office = Office.PowerPoint}},
            {"Create a presentation", new Intention()          {context= Context.Office, contextOnly= false, office = Office.PowerPoint}},
            {"Create a brief", new Intention()                 {context= Context.Office, contextOnly= false, office = Office.PowerPoint}},
            {"Next slide please", new Intention()              {context= Context.Office, contextOnly= false, office = Office.PPTNextSlide}},
            {"Previous slide please", new Intention()          {context= Context.Office, contextOnly= false, office = Office.PPTPreviousSlide}},
            {"Start the presentation", new Intention()         {context= Context.Office, contextOnly= false, office = Office.StartPresentation}},
            {"Word", new Intention()                           {context= Context.Office, contextOnly= false, office = Office.Word}},
            {"Open Word", new Intention()                      {context= Context.Office, contextOnly= false, office = Office.Word}},
            {"Create a document", new Intention()              {context= Context.Office, contextOnly= false, office = Office.Word}},
            {"Create a new document", new Intention()          {context= Context.Office, contextOnly= false, office = Office.Word}},
            {"Start a document", new Intention()               {context= Context.Office, contextOnly= false, office = Office.Word}},
            {"Start a new document", new Intention()           {context= Context.Office, contextOnly= false, office = Office.Word}},
            {"Open the document", new Intention()              {context= Context.Office, contextOnly= false, office = Office.OpenADocument}},
            {"Open the project plan", new Intention()          {context= Context.Office, contextOnly= false, office = Office.OpenAProjectPlan}},
        };

        Dictionary<string, Intention> OperationsPhrases = new Dictionary<string, Intention>()
        {
            {"Skeleton tracker", new Intention()                 {context= Context.Operations, contextOnly= false, kinnyops= Operations.SkeletonTracker}},
            {"Camera up", new Intention()                        {context= Context.Operations, contextOnly= false, kinnyops= Operations.CameraUp}},
            {"Lily please tilt the camera up", new Intention()   {context= Context.Operations, contextOnly= false, kinnyops= Operations.CameraUp}},
            {"Camera down", new Intention()                      {context= Context.Operations, contextOnly= false, kinnyops= Operations.CameraDown}},
            {"Lily please tilt the camera down", new Intention() {context= Context.Operations, contextOnly= false, kinnyops= Operations.CameraDown}},
            {"Camera center", new Intention()                    {context= Context.Operations, contextOnly= false, kinnyops= Operations.CameraCenter}},
            {"Lily please center the camera", new Intention()    {context= Context.Operations, contextOnly= false, kinnyops= Operations.CameraCenter}},
            {"course control", new Intention()                   {context= Context.Operations, contextOnly= false, kinnyops= Operations.CourseControl}},
            {"fine control", new Intention()                     {context= Context.Operations, contextOnly= false, kinnyops= Operations.FineControl}},
        };

        Dictionary<string, Intention> MeetingPhrases = new Dictionary<string, Intention>()
        {
            {"Record this", new Intention()                    {context= Context.Meeting, contextOnly= false, meeting= Meeting.CaptureAudio}},
            {"Video this", new Intention()                     {context= Context.Meeting, contextOnly= false, meeting= Meeting.CaptureVideo}},
            {"Stop recording", new Intention()                 {context= Context.Meeting, contextOnly= false, meeting= Meeting.Stop}},
            {"Lily snapshot", new Intention()                  {context= Context.Meeting, contextOnly= false, meeting= Meeting.Snapshot}},
            {"Lily please take a snapshot", new Intention()    {context= Context.Meeting, contextOnly= false, meeting= Meeting.Snapshot}},
            {"Lily take a snapshot", new Intention()           {context= Context.Meeting, contextOnly= false, meeting= Meeting.Snapshot}},
        };

        Dictionary<string, Intention> TranslationPhrases = new Dictionary<string, Intention>()
        {
            {"Arabic to English", new Intention()              {context= Context.Translate, contextOnly= false, translate= Translate.ArabicEnglish}},
            {"English to Arabic", new Intention()              {context= Context.Translate, contextOnly= false, translate= Translate.EnglishArabic}},
            {"English to Spanish", new Intention()             {context= Context.Translate, contextOnly= false, translate= Translate.EnglishSpanish}},
            {"English to Turkish", new Intention()             {context= Context.Translate, contextOnly= false, translate= Translate.EnglishTurkish}},
            {"Spanish to English", new Intention()             {context= Context.Translate, contextOnly= false, translate= Translate.SpanishEnglish}},
            {"Turkish to English", new Intention()             {context= Context.Translate, contextOnly= false, translate= Translate.TurkishEnglish}},
        };

        Dictionary<string, Intention> EntertainPhrases = new Dictionary<string, Intention>()
        {
            {"I want to listen to music", new Intention()      {context= Context.Entertain, contextOnly= false, entertain= Entertain.Music}},
            {"Can we listen to music", new Intention()         {context= Context.Entertain, contextOnly= false, entertain= Entertain.Music}},
            {"Lily, music please", new Intention()             {context= Context.Entertain, contextOnly= false, entertain= Entertain.Music}},
            {"I need some music", new Intention()              {context= Context.Entertain, contextOnly= false, entertain= Entertain.Music}},
            {"I want to watch a movie", new Intention()        {context= Context.Entertain, contextOnly= false, entertain= Entertain.Video}},
            {"Lily, movie please", new Intention()             {context= Context.Entertain, contextOnly= false, entertain= Entertain.Video}},
            {"Lets watch a movie", new Intention()             {context= Context.Entertain, contextOnly= false, entertain= Entertain.Video}},
            {"Can we watch a movie", new Intention()           {context= Context.Entertain, contextOnly= false, entertain= Entertain.Video}},
        };

        #endregion

        #region Class member variables

        private KinectAudioSource kinectSource;
        private RecognizerInfo ri;
        private SpeechRecognitionEngine sre;
        private const string RecognizerId = "SR_MS_en-US_Kinect_10.0";
        private bool keepListening = true;
        private List<Dictionary<string, Intention>> allDicts;
        private Grammar gContext;
        private Grammar gGeneral;
        private Grammar gResearch;
        private Grammar gOffice;
        private Grammar gOperations;
        private Grammar gMeeting;
        private Grammar gTranslation;
        private Grammar gEntertain;
       

        #endregion

        #region Constructors
        public SpeechTracker()
        {
            ri = SpeechRecognitionEngine.InstalledRecognizers().Where(r => r.Id == RecognizerId).FirstOrDefault();
            if (ri == null)
            {
                // no RecognizerInfo => bail
                return;
            }
            sre = new SpeechRecognitionEngine(ri.Id);
            // build our categories of Choices
            var contextPhrases = new Choices();
            foreach (var phrase in ContextPhrases)
                contextPhrases.Add(phrase.Key);

            var generalPhrases = new Choices();
            foreach (var phrase in GeneralPhrases)
                generalPhrases.Add(phrase.Key);

            var researchPhrases = new Choices();
            foreach (var phrase in ResearchPhrases)
                researchPhrases.Add(phrase.Key);

            var officePhrases = new Choices();
            foreach (var phrase in OfficePhrases)
                officePhrases.Add(phrase.Key);

            var operationsPhrases = new Choices();
            foreach (var phrase in OperationsPhrases)
                operationsPhrases.Add(phrase.Key);

            var meetingsPhrases = new Choices();
            foreach (var phrase in MeetingPhrases)
                meetingsPhrases.Add(phrase.Key);

            var translationPhrases = new Choices();
            foreach (var phrase in TranslationPhrases)
                translationPhrases.Add(phrase.Key);

            var entertainPhrases = new Choices();
            foreach (var phrase in EntertainPhrases)
                entertainPhrases.Add(phrase.Key);

            //create our GrammarBuilders
            var gbContext = new GrammarBuilder();
            gbContext.Append(contextPhrases);

            var gbGeneral = new GrammarBuilder();
            gbGeneral.Append(generalPhrases);

            var gbResearch = new GrammarBuilder();
            gbResearch.Append(researchPhrases);

            var gbOffice = new GrammarBuilder();
            gbOffice.Append(officePhrases);

            var gbOperations = new GrammarBuilder();
            gbOperations.Append(operationsPhrases);

            var gbMeeting = new GrammarBuilder();
            gbMeeting.Append(meetingsPhrases);

            var gbTranslation = new GrammarBuilder();
            gbTranslation.Append(translationPhrases);

            var gbEntertain = new GrammarBuilder();
            gbEntertain.Append(entertainPhrases);

            //and finally create our Grammars
            gContext = new Grammar(gbContext);
            gGeneral = new Grammar(gbGeneral);
            gResearch = new Grammar(gbResearch);
            gOffice = new Grammar(gbOffice);
            gOperations = new Grammar(gbOperations);
            gMeeting = new Grammar(gbMeeting);
            gTranslation = new Grammar(gbTranslation);
            gEntertain = new Grammar(gbEntertain);

            //we are only going to load the Context and General grammars at this point
            sre.LoadGrammar(gContext);
            sre.LoadGrammar(gGeneral);

            allDicts = new List<Dictionary<string, Intention>>() { ContextPhrases, 
                                                                   GeneralPhrases, 
                                                                   ResearchPhrases, 
                                                                   OfficePhrases, 
                                                                   OperationsPhrases, 
                                                                   MeetingPhrases, 
                                                                   TranslationPhrases, 
                                                                   EntertainPhrases };

            sre.SpeechRecognized += new EventHandler<SpeechRecognizedEventArgs>(sre_SpeechRecognized);
            sre.SpeechHypothesized += new EventHandler<SpeechHypothesizedEventArgs>(sre_SpeechHypothesized);
            sre.SpeechRecognitionRejected += new EventHandler<SpeechRecognitionRejectedEventArgs>(sre_SpeechRecognitionRejected);

            var t = new Thread(BeginListening);
            t.Priority = ThreadPriority.Highest;
            t.Start();
        }

        void sre_SpeechRecognitionRejected(object sender, SpeechRecognitionRejectedEventArgs e)
        {
            //throw new NotImplementedException();
        }

        void sre_SpeechHypothesized(object sender, SpeechHypothesizedEventArgs e)
        {
            //throw new NotImplementedException();
        }

        void sre_SpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {
            if (e.Result.Confidence < .95f)
            {
                return;
            }

            if (SpeechDetected == null)
                return;

            var intent = new IntentionArgs();

            bool found = false;
            for (int i = 0; i < allDicts.Count && !found; i++)
            {
                foreach (var phrase in allDicts[i])
                {
                    if (e.Result.Text.Equals(phrase.Key))
                    {
                        //we have a match from our grammars => process the audio command
                        intent.context = phrase.Value.context;
                        intent.entertain = phrase.Value.entertain;
                        intent.general = phrase.Value.general;
                        intent.operations = phrase.Value.kinnyops;
                        intent.meeting = phrase.Value.meeting;
                        intent.office = phrase.Value.office;
                        intent.research = phrase.Value.research;
                        intent.translate = phrase.Value.translate;
                        intent.contextOnly = phrase.Value.contextOnly;
                        intent.phrase = e.Result.Text;
                        found = true;
                        break;
                    }
                }
            }
            if (!found)
            {
                // the speech detected was not matched => continue listening
                return;
            }
            if (intent.contextOnly)
            {
                //this command is only switching context => we do not need to raise the SpeechDetected event; swap out grammars based on context
                //however, we want to throw the SpeechContextChanged event if it is being listened to
                if (SpeechContextChanged != null)
                {
                    SpeechContextChangedArgs arg = new SpeechContextChangedArgs();
                    arg.context = intent.context;
                    SpeechContextChanged(this, arg);
                }
                switch (intent.context)
                {
                    case Context.Entertain:
                        ClearGrammars();
                        sre.LoadGrammar(gEntertain);
                        break;
                    case Context.Operations:
                        ClearGrammars();
                        sre.LoadGrammar(gOperations);
                        break;
                    case Context.Meeting:
                        ClearGrammars();
                        sre.LoadGrammar(gMeeting);
                        break;
                    case Context.Office:
                        ClearGrammars();
                        sre.LoadGrammar(gOffice);
                        break;
                    case Context.Research:
                        ClearGrammars();
                        sre.LoadGrammar(gResearch);
                        break;
                    case Context.Translate:
                        GeneralActions.SpeakAsync("Yes. I can translate for you. Tell me the languages you want translated from, and to.");
                        ClearGrammars();
                        sre.LoadGrammar(gTranslation);
                        break;
                    default:
                        break;
                }                
            }
            else
            {
                SpeechDetected(this, intent);
            }
        }

        private void ClearGrammars()
        {
            List<Grammar> remove = new List<Grammar>();
            foreach (Grammar g in sre.Grammars)
            {
                if (!(g.Equals(gContext) || g.Equals(gGeneral)))
                {
                    remove.Add(g);
                }
            }
            foreach (Grammar g in remove)
            {
                sre.UnloadGrammar(g);
            }
        }

        #endregion

        #region Private methods

        private void BeginListening()
        {
            kinectSource = new KinectAudioSource();
            kinectSource.SystemMode = SystemMode.OptibeamArrayOnly;
            kinectSource.FeatureMode = true;
            kinectSource.AutomaticGainControl = true;
            kinectSource.NoiseSuppression = true;
            kinectSource.MicArrayMode = MicArrayMode.MicArrayAdaptiveBeam;
            var kinectStream = kinectSource.Start();
            sre.SetInputToAudioStream(kinectStream,
                                      new SpeechAudioFormatInfo(EncodingFormat.Pcm,
                                                                              16000,
                                                                              16,
                                                                              1,
                                                                              32000,
                                                                              2,
                                                                              null));
            sre.RecognizeAsync(RecognizeMode.Multiple);
        }

        private void StopListening()
        {
            if (sre != null)
            {
                sre.RecognizeAsyncCancel();
                sre.RecognizeAsyncStop();
                kinectSource.Dispose();
            }
        }

        #endregion

        #region Protected methods



        #endregion

        #region Public methods



        #endregion

        #region Properties

        public bool KeepListening
        {
            get { return keepListening; }
            set
            {
                keepListening = value;
                if (!keepListening)
                {
                    // it is no longer desired for Lily to be listening => 
                    StopListening();
                }
            }
        }

        #endregion


    }
}
